/*
 *  linux/fs/xip2/super.c, Version 1
 *
 * (C) Copyright IBM Corp. 2002,2004
 * Author(s): Carsten Otte <cotte@de.ibm.com>
 * derived from second extended filesystem (ext2)
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <linux/xip2_fs.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/locks.h>
#include <linux/blkdev.h>
#include <asm/dcss.h>
#include <asm/uaccess.h>

static char error_buf[1024];

void xip2_error (struct super_block * sb, const char * function,
		 const char * fmt, ...)
{
	va_list args;

	va_start (args, fmt);
	vsprintf (error_buf, fmt, args);
	va_end (args);
	if (test_opt (sb, ERRORS_PANIC) ||
	    (le16_to_cpu(sb->u.xip2_sb.s_es->s_errors) == EXT2_ERRORS_PANIC &&
	     !test_opt (sb, ERRORS_CONT) && !test_opt (sb, ERRORS_RO)))
		panic ("XIP2-fs panic (device %s): %s: %s\n",
		       bdevname(sb->s_dev), function, error_buf);
	printk (KERN_CRIT "XIP2-fs error (device %s): %s: %s\n",
		bdevname(sb->s_dev), function, error_buf);
	if (test_opt (sb, ERRORS_RO) ||
	    (le16_to_cpu(sb->u.xip2_sb.s_es->s_errors) == EXT2_ERRORS_RO &&
	     !test_opt (sb, ERRORS_CONT) && !test_opt (sb, ERRORS_PANIC))) {
		printk ("Remounting filesystem read-only\n");
		sb->s_flags |= MS_RDONLY;
	}
}

NORET_TYPE void xip2_panic (struct super_block * sb, const char * function,
			    const char * fmt, ...)
{
	va_list args;

	if (!(sb->s_flags & MS_RDONLY)) {
		sb->u.xip2_sb.s_mount_state |= EXT2_ERROR_FS;
		sb->u.xip2_sb.s_es->s_state =
			cpu_to_le16(le16_to_cpu(sb->u.xip2_sb.s_es->s_state) 
				    | EXT2_ERROR_FS);
		sb->s_dirt = 1;
	}
	va_start (args, fmt);
	vsprintf (error_buf, fmt, args);
	va_end (args);
	sb->s_flags |= MS_RDONLY;
	panic ("XIP2-fs panic (device %s): %s: %s\n",
	       bdevname(sb->s_dev), function, error_buf);
}

void xip2_warning (struct super_block * sb, const char * function,
		   const char * fmt, ...)
{
	va_list args;

	va_start (args, fmt);
	vsprintf (error_buf, fmt, args);
	va_end (args);
	printk (KERN_WARNING "XIP2-fs warning (device %s): %s: %s\n",
		bdevname(sb->s_dev), function, error_buf);
}

void xip2_update_dynamic_rev(struct super_block *sb)
{
	struct ext2_super_block *es = XIP2_SB(sb)->s_es;

	if (le32_to_cpu(es->s_rev_level) > EXT2_GOOD_OLD_REV)
		return;

	xip2_warning(sb, __FUNCTION__,
		     "updating to rev %d because of new feature flag, "
		     "running e2fsck is recommended",
		     EXT2_DYNAMIC_REV);

	es->s_first_ino = cpu_to_le32(EXT2_GOOD_OLD_FIRST_INO);
	es->s_inode_size = cpu_to_le16(EXT2_GOOD_OLD_INODE_SIZE);
	es->s_rev_level = cpu_to_le32(EXT2_DYNAMIC_REV);
	/* leave es->s_feature_*compat flags alone */
	/* es->s_uuid will be set by e2fsck if empty */

	/*
	 * The rest of the superblock fields should be zero, and if not it
	 * means they are likely already in use, so leave them alone.  We
	 * can leave it up to e2fsck to clean up any inconsistencies there.
	 */
}

void xip2_put_super (struct super_block * sb)
{
	if (sb->u.xip2_sb.mem_area) {
		segment_unload (((xip2_mem_area_t*)(sb->u.xip2_sb.mem_area))
				->name);
		kfree (sb->u.xip2_sb.mem_area);
	}
	kfree(sb->u.xip2_sb.s_group_desc);
	return;
}

static struct super_operations xip2_sops = {
	read_inode:	xip2_read_inode,
	put_super:	xip2_put_super,
	statfs:		xip2_statfs,
	remount_fs:	xip2_remount,
};

/*
 * This function has been shamelessly adapted from the msdos fs
 */
static int parse_options (char * options, unsigned long * sb_block,
			  unsigned short *resuid, unsigned short * resgid,
			  unsigned long * mount_options,
			  xip2_mem_area_t ** mem_area)
{
	char * this_char;
	char * value;

	if (!options)
		return 1;
	for (this_char = strtok (options, ",");
	     this_char != NULL;
	     this_char = strtok (NULL, ",")) {
		if ((value = strchr (this_char, '=')) != NULL)
			*value++ = 0;
		if (!strcmp (this_char, "bsddf"))
			clear_opt (*mount_options, MINIX_DF);
		else if (!strcmp (this_char, "nouid32")) {
			set_opt (*mount_options, NO_UID32);
		}
		else if (!strcmp (this_char, "check")) {
			if (!value || !*value || !strcmp (value, "none"))
				clear_opt (*mount_options, CHECK);
			else
#ifdef CONFIG_EXT2_CHECK
				set_opt (*mount_options, CHECK);
#else
				printk("XIP2 Check option not supported\n");
#endif
		}
		else if (!strcmp (this_char, "debug"))
			set_opt (*mount_options, DEBUG);
		else if (!strcmp (this_char, "errors")) {
			if (!value || !*value) {
				printk ("XIP2-fs: the errors option requires "
					"an argument\n");
				return 0;
			}
			if (!strcmp (value, "continue")) {
				clear_opt (*mount_options, ERRORS_RO);
				clear_opt (*mount_options, ERRORS_PANIC);
				set_opt (*mount_options, ERRORS_CONT);
			}
			else if (!strcmp (value, "remount-ro")) {
				clear_opt (*mount_options, ERRORS_CONT);
				clear_opt (*mount_options, ERRORS_PANIC);
				set_opt (*mount_options, ERRORS_RO);
			}
			else if (!strcmp (value, "panic")) {
				clear_opt (*mount_options, ERRORS_CONT);
				clear_opt (*mount_options, ERRORS_RO);
				set_opt (*mount_options, ERRORS_PANIC);
			}
			else {
				printk ("XIP2-fs: Invalid errors option: %s\n",
					value);
				return 0;
			}
		}
		else if (!strcmp (this_char, "grpid") ||
			 !strcmp (this_char, "bsdgroups"))
			set_opt (*mount_options, GRPID);
		else if (!strcmp (this_char, "minixdf"))
			set_opt (*mount_options, MINIX_DF);
		else if (!strcmp (this_char, "nocheck"))
			clear_opt (*mount_options, CHECK);
		else if (!strcmp (this_char, "nogrpid") ||
			 !strcmp (this_char, "sysvgroups"))
			clear_opt (*mount_options, GRPID);
		else if (!strcmp (this_char, "resgid")) {
			if (!value || !*value) {
				printk ("XIP2-fs: the resgid option requires "
					"an argument\n");
				return 0;
			}
			*resgid = simple_strtoul (value, &value, 0);
			if (*value) {
				printk ("XIP2-fs: Invalid resgid option: %s\n",
					value);
				return 0;
			}
		}
		else if (!strcmp (this_char, "resuid")) {
			if (!value || !*value) {
				printk ("XIP2-fs: the resuid option requires "
					"an argument");
				return 0;
			}
			*resuid = simple_strtoul (value, &value, 0);
			if (*value) {
				printk ("XIP2-fs: Invalid resuid option: %s\n",
					value);
				return 0;
			}
		}
		else if (!strcmp (this_char, "sb")) {
			if (!value || !*value) {
				printk ("XIP2-fs: the sb option requires "
					"an argument");
				return 0;
			}
			*sb_block = simple_strtoul (value, &value, 0);
			if (*value) {
				printk ("XIP2-fs: Invalid sb option: %s\n",
					value);
				return 0;
			}
		}
		else if (!strcmp (this_char, "memarea")) {
			if (!value || !*value) {
				printk ("XIP2-fs: the memarea option requires "
					"an argument\n");
				return 0;
			}
			*mem_area=kmalloc (sizeof(xip2_mem_area_t)+strlen(value)+1,GFP_ATOMIC);
			if (mem_area==NULL) {
				printk ("XIP2-fs: out of memory\n");
				return 0;
			}
			(*mem_area)->name=((char*)(*mem_area)) +
					sizeof(xip2_mem_area_t);
			strcpy ((*mem_area)->name, value);
			if (segment_load(value, SEGMENT_SHARED_RO, &(*mem_area)->start, 
					 &(*mem_area)->end)<0) {
				printk ("XIP2-fs: the memarea specified could "
					"not be found\n");
				kfree (*mem_area);
				*mem_area = NULL;
				return 0;
			}
		}
		/* Silently ignore the quota options */
		else if (!strcmp (this_char, "grpquota")
		         || !strcmp (this_char, "noquota")
		         || !strcmp (this_char, "quota")
		         || !strcmp (this_char, "usrquota"))
			/* Don't do anything ;-) */ ;
		else {
			printk ("XIP2-fs: Unrecognized mount option %s\n", 
				this_char);
			return 0;
		}
	}
	return 1;
}

static int xip2_check_descriptors (struct super_block * sb)
{
	int i;
	int desc_block = 0;
	unsigned long block = le32_to_cpu(sb->u.xip2_sb.s_es->
					  s_first_data_block);
	struct ext2_group_desc * gdp = NULL;

	xip2_debug ("Checking group descriptors");

	for (i = 0; i < sb->u.xip2_sb.s_groups_count; i++)
	{
		if ((i % EXT2_DESC_PER_BLOCK(sb)) == 0)
			gdp = (struct ext2_group_desc *) 
				sb->u.xip2_sb.s_group_desc[desc_block++];
		if (le32_to_cpu(gdp->bg_block_bitmap) < block ||
		    le32_to_cpu(gdp->bg_block_bitmap) >= 
		    block + EXT2_BLOCKS_PER_GROUP(sb))
		{
			xip2_error (sb, "xip2_check_descriptors",
				    "Block bitmap for group %d"
				    " not in group (block %lu)!",
				    i, (unsigned long) 
				    le32_to_cpu(gdp->bg_block_bitmap));
			return 0;
		}
		if (le32_to_cpu(gdp->bg_inode_bitmap) < block ||
		    le32_to_cpu(gdp->bg_inode_bitmap) >= block + 
		    EXT2_BLOCKS_PER_GROUP(sb))
		{
			xip2_error (sb, "xip2_check_descriptors",
				    "Inode bitmap for group %d"
				    " not in group (block %lu)!",
				    i, (unsigned long) 
				    le32_to_cpu(gdp->bg_inode_bitmap));
			return 0;
		}
		if (le32_to_cpu(gdp->bg_inode_table) < block ||
		    le32_to_cpu(gdp->bg_inode_table) + 
		    sb->u.xip2_sb.s_itb_per_group >=
		    block + EXT2_BLOCKS_PER_GROUP(sb))
		{
			xip2_error (sb, "xip2_check_descriptors",
				    "Inode table for group %d"
				    " not in group (block %lu)!",
				    i, (unsigned long) 
				    le32_to_cpu(gdp->bg_inode_table));
			return 0;
		}
		block += EXT2_BLOCKS_PER_GROUP(sb);
		gdp++;
	}
	return 1;
}

#define log2(n) ffz(~(n))
 
/*
 * Maximal file size.  There is a direct, and {,double-,triple-}indirect
 * block limit, and also a limit of (2^32 - 1) 512-byte sectors in i_blocks.
 * We need to be 1 filesystem block less than the 2^32 sector limit.
 */
static loff_t xip2_max_size(int bits)
{
	loff_t res = EXT2_NDIR_BLOCKS;
	res += 1LL << (bits-2);
	res += 1LL << (2*(bits-2));
	res += 1LL << (3*(bits-2));
	res <<= bits;
	if (res > (512LL << 32) - (1 << bits))
		res = (512LL << 32) - (1 << bits);
	return res;
}

struct super_block * xip2_read_super (struct super_block * sb, void * data,
				      int silent)
{
	xip2_mem_area_t* mem_area = NULL;
	void * block_ptr;
	struct ext2_super_block * es;
	unsigned long sb_block = 0;
	unsigned short resuid = EXT2_DEF_RESUID;
	unsigned short resgid = EXT2_DEF_RESGID;
	kdev_t dev = sb->s_dev;
	int blocksize = BLOCK_SIZE;
	int db_count;
	int i;


	blocksize = PAGE_SIZE;
	sb->u.xip2_sb.s_mount_opt = 0;
	if (!parse_options ((char *) data, &sb_block, &resuid, &resgid,
	    &sb->u.xip2_sb.s_mount_opt,&mem_area)) {
		return NULL;
	}

	if (!mem_area) {
		printk ("XIP2-fs: no memarea specified\n");
		return NULL;
	}

	if (!(block_ptr = xip2_maread (mem_area, sb_block, blocksize))) {
		printk ("XIP2-fs: unable to read superblock\n");
		kfree (mem_area);
		return NULL;
	}
	/*
	 * Note: s_es must be initialized as soon as possible because
	 *       some ext2 macro-instructions depend on its value
	 */
	es = (struct ext2_super_block *) (block_ptr + BLOCK_SIZE);
	sb->u.xip2_sb.s_es = es;
	sb->u.xip2_sb.mem_area = mem_area;
	sb->s_magic = le16_to_cpu(es->s_magic);
	if (sb->s_magic != EXT2_SUPER_MAGIC) {
		if (!silent)
			printk ("VFS: Can't find xip2 filesystem on "
				"memarea %s.\n", mem_area->name);
		goto failed_mount;
	}
	if (le32_to_cpu(es->s_rev_level) == EXT2_GOOD_OLD_REV &&
	    (EXT2_HAS_COMPAT_FEATURE(sb, ~0U) ||
	     EXT2_HAS_RO_COMPAT_FEATURE(sb, ~0U) ||
	     EXT2_HAS_INCOMPAT_FEATURE(sb, ~0U)))
		printk("XIP2-fs warning: feature flags set on rev 0 fs, "
		       "running e2fsck is recommended\n");
	/*
	 * Check if fs should be mounted rdonly. If not let the mount fail.
	 */
	if (!(sb->s_flags & MS_RDONLY)) {
		printk("XIP2-fs: %s: mounting RDWR was requested, "
		       "but is not supported. Please mount using"
		       "the RO option\n", mem_area->name);
		goto failed_mount;
	}
	
	/*
	 * Check feature flags regardless of the revision level, since we
	 * previously didn't change the revision level when setting the flags,
	 * so there is a chance incompat flags are set on a rev 0 filesystem.
	 */
	if ((i = EXT2_HAS_INCOMPAT_FEATURE(sb, ~EXT2_FEATURE_INCOMPAT_SUPP))) {
		printk("XIP2-fs: %s: couldn't mount because of "
		       "unsupported optional features (%x).\n",
		       mem_area->name, i);
		goto failed_mount;
	}
	sb->s_blocksize_bits =
		le32_to_cpu(XIP2_SB(sb)->s_es->s_log_block_size) + 10;
	sb->s_blocksize = 1 << sb->s_blocksize_bits;

	sb->s_maxbytes = xip2_max_size(sb->s_blocksize_bits);

	/* If the blocksize doesn't match, tell the user.. */
	if (sb->s_blocksize != blocksize) {
		printk("XIP2-fs: Blocksize of the filesystem "
		       "on %s does not match PAGE_SIZE.\n",
		       mem_area->name);
			goto failed_mount;
		}

	if (le32_to_cpu(es->s_rev_level) == EXT2_GOOD_OLD_REV) {
		sb->u.xip2_sb.s_inode_size = EXT2_GOOD_OLD_INODE_SIZE;
		sb->u.xip2_sb.s_first_ino = EXT2_GOOD_OLD_FIRST_INO;
	} else {
		sb->u.xip2_sb.s_inode_size = le16_to_cpu(es->s_inode_size);
		sb->u.xip2_sb.s_first_ino = le32_to_cpu(es->s_first_ino);
		if (sb->u.xip2_sb.s_inode_size != EXT2_GOOD_OLD_INODE_SIZE) {
			printk ("XIP2-fs: unsupported inode size: %d\n",
				sb->u.xip2_sb.s_inode_size);
			goto failed_mount;
		}
	}
	sb->u.xip2_sb.s_frag_size = EXT2_MIN_FRAG_SIZE <<
				   le32_to_cpu(es->s_log_frag_size);
	if (sb->u.xip2_sb.s_frag_size)
		sb->u.xip2_sb.s_frags_per_block = sb->s_blocksize /
						  sb->u.xip2_sb.s_frag_size;
	else
		sb->s_magic = 0;
	sb->u.xip2_sb.s_blocks_per_group = le32_to_cpu(es->s_blocks_per_group);
	sb->u.xip2_sb.s_frags_per_group = le32_to_cpu(es->s_frags_per_group);
	sb->u.xip2_sb.s_inodes_per_group = le32_to_cpu(es->s_inodes_per_group);
	sb->u.xip2_sb.s_inodes_per_block = sb->s_blocksize /
					   EXT2_INODE_SIZE(sb);
	sb->u.xip2_sb.s_itb_per_group = sb->u.xip2_sb.s_inodes_per_group /
				        sb->u.xip2_sb.s_inodes_per_block;
	sb->u.xip2_sb.s_desc_per_block = sb->s_blocksize /
					 sizeof (struct ext2_group_desc);
	sb->u.xip2_sb.s_sbp = block_ptr;
	if (resuid != EXT2_DEF_RESUID)
		sb->u.xip2_sb.s_resuid = resuid;
	else
		sb->u.xip2_sb.s_resuid = le16_to_cpu(es->s_def_resuid);
	if (resgid != EXT2_DEF_RESGID)
		sb->u.xip2_sb.s_resgid = resgid;
	else
		sb->u.xip2_sb.s_resgid = le16_to_cpu(es->s_def_resgid);
	sb->u.xip2_sb.s_mount_state = le16_to_cpu(es->s_state);
	sb->u.xip2_sb.s_addr_per_block_bits =
		log2 (EXT2_ADDR_PER_BLOCK(sb));
	sb->u.xip2_sb.s_desc_per_block_bits =
		log2 (EXT2_DESC_PER_BLOCK(sb));
	if (sb->s_magic != EXT2_SUPER_MAGIC) {
		if (!silent)
			printk ("VFS: Can't find an xip2 filesystem on dev "
				"%s.\n",
				bdevname(dev));
		goto failed_mount;
	}

	if (sb->s_blocksize != sb->u.xip2_sb.s_frag_size) {
		printk ("XIP2-fs: fragsize %lu != blocksize %lu "
			"(not supported yet)\n",
			sb->u.xip2_sb.s_frag_size, sb->s_blocksize);
		goto failed_mount;
	}

	if (sb->u.xip2_sb.s_blocks_per_group > sb->s_blocksize * 8) {
		printk ("XIP2-fs: #blocks per group too big: %lu\n",
			sb->u.xip2_sb.s_blocks_per_group);
		goto failed_mount;
	}
	if (sb->u.xip2_sb.s_frags_per_group > sb->s_blocksize * 8) {
		printk ("XIP2-fs: #fragments per group too big: %lu\n",
			sb->u.xip2_sb.s_frags_per_group);
		goto failed_mount;
	}
	if (sb->u.xip2_sb.s_inodes_per_group > sb->s_blocksize * 8) {
		printk ("XIP2-fs: #inodes per group too big: %lu\n",
			sb->u.xip2_sb.s_inodes_per_group);
		goto failed_mount;
	}

	sb->u.xip2_sb.s_groups_count = (le32_to_cpu(es->s_blocks_count) -
				        le32_to_cpu(es->s_first_data_block) +
				       EXT2_BLOCKS_PER_GROUP(sb) - 1) /
				       EXT2_BLOCKS_PER_GROUP(sb);
	db_count = (sb->u.xip2_sb.s_groups_count + EXT2_DESC_PER_BLOCK(sb) - 1)
		/ EXT2_DESC_PER_BLOCK(sb);
	sb->u.xip2_sb.s_group_desc = kmalloc (db_count * sizeof (void *), 
					      GFP_KERNEL);
	if (sb->u.xip2_sb.s_group_desc == NULL) {
		printk ("XIP2-fs: not enough memory\n");
		goto failed_mount;
	}
	for (i = 0; i < db_count; i++) {
		sb->u.xip2_sb.s_group_desc[i] = xip2_maread 
			(mem_area, sb_block + i + 1, sb->s_blocksize);
		if (!sb->u.xip2_sb.s_group_desc[i]) {
			kfree(sb->u.xip2_sb.s_group_desc);
			printk ("XIP2-fs: unable to read group descriptors\n");
			goto failed_mount;
		}
	}
	if (!xip2_check_descriptors (sb)) {
		printk ("XIP2-fs: group descriptors corrupted!\n");
		db_count = i;
		goto failed_mount2;
	}
	for (i = 0; i < EXT2_MAX_GROUP_LOADED; i++) {
		sb->u.xip2_sb.s_inode_bitmap_number[i] = 0;
		sb->u.xip2_sb.s_inode_bitmap[i] = NULL;
		sb->u.xip2_sb.s_block_bitmap_number[i] = 0;
		sb->u.xip2_sb.s_block_bitmap[i] = NULL;
	}
	sb->u.xip2_sb.s_loaded_inode_bitmaps = 0;
	sb->u.xip2_sb.s_loaded_block_bitmaps = 0;
	sb->u.xip2_sb.s_gdb_count = db_count;
	/*
	 * set up enough so that it can read an inode
	 */
	sb->s_op = &xip2_sops;
	sb->s_root = d_alloc_root(iget(sb, EXT2_ROOT_INO));
	if (!sb->s_root || !S_ISDIR(sb->s_root->d_inode->i_mode) ||
	    !sb->s_root->d_inode->i_blocks || !sb->s_root->d_inode->i_size) {
		if (sb->s_root) {
			dput(sb->s_root);
			sb->s_root = NULL;
			printk(KERN_ERR "XIP2-fs: corrupt root inode, run "
			       "e2fsck\n");
		} else
			printk(KERN_ERR "XIP2-fs: get root inode failed\n");
		goto failed_mount2;
	}
	return sb;
failed_mount2:
	kfree(sb->u.xip2_sb.s_group_desc);
failed_mount:
	kfree(mem_area);
	return NULL;
}

int xip2_remount (struct super_block * sb, int * flags, char * data)
{
	struct ext2_super_block * es;
	xip2_mem_area_t* mem_area;
	unsigned short resuid = sb->u.xip2_sb.s_resuid;
	unsigned short resgid = sb->u.xip2_sb.s_resgid;
	unsigned long new_mount_opt;
	unsigned long tmp;

	/* filesystem is read-only _only_, therefore mounting rdwr is not permitted */
	if (!(*flags & MS_RDONLY)) 
		return -EINVAL;

	/*
	 * Allow the "check" option to be passed as a remount option.
	 */
	new_mount_opt = sb->u.xip2_sb.s_mount_opt;
	mem_area = sb->u.xip2_sb.mem_area;
	if (!parse_options (data, &tmp, &resuid, &resgid,
			    &new_mount_opt, &mem_area))
		return -EINVAL;

	if ((!mem_area) || (mem_area != sb->u.xip2_sb.mem_area))
		return -EINVAL;

	sb->u.xip2_sb.s_mount_opt = new_mount_opt;
	sb->u.xip2_sb.s_resuid = resuid;
	sb->u.xip2_sb.s_resgid = resgid;
	es = sb->u.xip2_sb.s_es;
	return 0;
}

int xip2_statfs (struct super_block * sb, struct statfs * buf)
{
	unsigned long overhead;
	int i;

	if (test_opt (sb, MINIX_DF))
		overhead = 0;
	else {
		/*
		 * Compute the overhead (FS structures)
		 */

		/*
		 * All of the blocks before first_data_block are
		 * overhead
		 */
		overhead = le32_to_cpu(sb->u.xip2_sb.s_es->s_first_data_block);

		/*
		 * Add the overhead attributed to the superblock and
		 * block group descriptors.  If the sparse superblocks
		 * feature is turned on, then not all groups have this.
		 */
		for (i = 0; i < XIP2_SB(sb)->s_groups_count; i++)
			overhead += xip2_bg_has_super(sb, i) +
				xip2_bg_num_gdb(sb, i);

		/*
		 * Every block group has an inode bitmap, a block
		 * bitmap, and an inode table.
		 */
		overhead += (sb->u.xip2_sb.s_groups_count *
			     (2 + sb->u.xip2_sb.s_itb_per_group));
	}

	buf->f_type = EXT2_SUPER_MAGIC;
	buf->f_bsize = sb->s_blocksize;
	buf->f_blocks = le32_to_cpu(sb->u.xip2_sb.s_es->s_blocks_count) 
		- overhead;
	buf->f_bfree = xip2_count_free_blocks (sb);
	buf->f_bavail = buf->f_bfree - 
		le32_to_cpu(sb->u.xip2_sb.s_es->s_r_blocks_count);
	if (buf->f_bfree < le32_to_cpu(sb->u.xip2_sb.s_es->s_r_blocks_count))
		buf->f_bavail = 0;
	buf->f_files = le32_to_cpu(sb->u.xip2_sb.s_es->s_inodes_count);
	buf->f_ffree = xip2_count_free_inodes (sb);
	buf->f_namelen = EXT2_NAME_LEN;
	return 0;
}

static DECLARE_FSTYPE (xip2_fs_type, "xip2", xip2_read_super, 
		       FS_NO_DCACHE|FS_NO_PRELIM);

static int __init init_xip2_fs(void)
{
        return register_filesystem(&xip2_fs_type);
}

static void __exit exit_xip2_fs(void)
{
	unregister_filesystem(&xip2_fs_type);
}

EXPORT_NO_SYMBOLS;

module_init(init_xip2_fs)
module_exit(exit_xip2_fs)
